﻿using System;
using System.Collections.Generic;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration {

  public abstract class Symbol : IEquatable<Symbol> {
    private readonly int stringRepresentationHash;

    public string StringRepresentation { get; }

    protected Symbol(string representation) {
      StringRepresentation = representation;
      stringRepresentationHash = representation.GetHashCode();
    }

    public override string ToString() {
      return StringRepresentation;
    }

    #region IEquatable
    public static bool operator ==(Symbol s1, Symbol s2) {
      return !ReferenceEquals(null, s1) && !ReferenceEquals(null, s2) && (ReferenceEquals(s1, s2)
        || s1.Equals(s2));
    }

    public static bool operator !=(Symbol s1, Symbol s2) {
      return !(s1 == s2);
    }

    public bool Equals(Symbol other) {
      if (ReferenceEquals(null, other)) return false;
      if (ReferenceEquals(this, other)) return true;
      return string.Equals(StringRepresentation, other.StringRepresentation);
    }

    public override bool Equals(object obj) {
      if (ReferenceEquals(null, obj)) return false;
      if (ReferenceEquals(this, obj)) return true;
      if (obj.GetType() != this.GetType()) return false;
      return Equals((Symbol)obj);
    }

    public override int GetHashCode() {
      return stringRepresentationHash;
    }
    #endregion
  }

  public class TerminalSymbol : Symbol {
    public readonly bool IsVariable;

    public TerminalSymbol(string representation, bool isVariable = false) : base(representation) {
      IsVariable = isVariable;
    }
  }

  public class NonterminalSymbol : Symbol {
    public List<Production> Alternatives { get; }

    public NonterminalSymbol(string representation) : base(representation) {
      Alternatives = new List<Production>();
    }

    public void AddProduction(params Symbol[] production) {
      Alternatives.Add(new Production(production));
    }
  }

  public class VariableSymbol : NonterminalSymbol { // Convenience class
    public IEnumerable<TerminalSymbol> VariableTerminalSymbols { get; }

    public VariableSymbol(string representation, IEnumerable<string> variableNames) : base(representation) {
      List<TerminalSymbol> createdSymbols = new List<TerminalSymbol>();
      VariableTerminalSymbols = createdSymbols;

      foreach (string variableName in variableNames) {
        TerminalSymbol s = new TerminalSymbol(variableName, isVariable: true);
        createdSymbols.Add(s);
        AddProduction(s);
      }
    }
  }

  public class Production : List<Symbol> {

    public Production(params Symbol[] symbols) : base(symbols) { }

    public Production(IEnumerable<Symbol> symbols) : base(symbols) { }

    public override string ToString() {
      return string.Join(" ", this);
    }
  }
}
